import pytest
from pytest_mock import mocker

from enum import Enum

from hangouts_helper.handler import HangoutsChatHandler, EventType, SpaceType


@pytest.fixture
def handler(mocker):
    class ActionMethod(Enum):
        TEST_ACTION = 'TEST_ACTION'
    handler = HangoutsChatHandler()
    handler.ActionMethod = ActionMethod
    return handler

@pytest.fixture
def create_event():

    def _create_event(event_type='MESSAGE', space_type='DM',
                      user_type='HUMAN', message_text=None,
                      action_method=None, action_parameters=None):
        event = {
            "type": event_type,
            "eventTime": '2018-08-04T01:36:33.832895Z',
            "token": 'yxsTv7uVsnr5BL1qecnig8sMLLulF5RvC0nTrw2yZCE=',
            "responseUrl": 'http://www.example.com',
            "threadKey": 'thread1',
            "configCompleteRedirectUrl": 'https://chat.google.com/api/bot_config_complete?token=abc123',
            "user": {
                "name": 'users/111998962929502904878',
                "displayName": 'Bob Dylan',
                "type": user_type
            },
            "space": {
                "name": 'spaces/AAAA9UbUwi8',
                "type": space_type,
                "displayName": 'My Chat'
            }
        }
        if message_text is not None:
            event.update({'message': {'text': message_text}})
        if action_method is not None:
            action_data = {'actionMethodName': action_method}
            if action_parameters is not None:
                action_data.update({'parameters': action_parameters})
            event.update({'action': action_data})
        return event
    return _create_event

@pytest.mark.parametrize("space_type", [SpaceType.ROOM, SpaceType.DIRECT_MESSAGE])
def test_handle_added_to_space(mocker, handler, create_event, space_type):
    event = create_event(event_type='ADDED_TO_SPACE', space_type=space_type.value)
    mocker.patch.object(HangoutsChatHandler, 'handle_added_to_space')
    handler.handle_chat_event(event)
    handler.handle_added_to_space.assert_called_once_with(space_type, event)

@pytest.mark.parametrize("space_type", [SpaceType.ROOM, SpaceType.DIRECT_MESSAGE])
def test_handle_removed_from_space(mocker, handler, create_event, space_type):
    space_type_str = space_type.value  # create_event expects string value
    event = create_event(event_type='REMOVED_FROM_SPACE', space_type=space_type_str)
    mocker.patch.object(HangoutsChatHandler, 'handle_removed_from_space')
    handler.handle_chat_event(event)
    handler.handle_removed_from_space.assert_called_once_with(space_type, event)

def test_handle_message(mocker, handler, create_event):
    event = create_event(event_type='MESSAGE', message_text='Hello')
    mocker.patch.object(HangoutsChatHandler, 'handle_message')
    handler.handle_chat_event(event)
    handler.handle_message.assert_called_once_with('Hello', event)

def test_handle_card_clicked(mocker, handler, create_event):
    action_method_name = 'TEST_ACTION'
    action_parameters = [
        {'key': 'key1', 'value': 'value1'},
        {'key': 'key2', 'value': 'value2'}
    ]
    event = create_event(event_type='CARD_CLICKED', action_method=action_method_name,
                         action_parameters=action_parameters)
    mocker.patch.object(HangoutsChatHandler, 'handle_card_clicked')
    handler.handle_chat_event(event)
    expected_parameters = {
        'key1': 'value1',
        'key2': 'value2'
    }
    handler.handle_card_clicked.assert_called_once_with(
        handler.ActionMethod.TEST_ACTION,
        expected_parameters,
        event)

def test_parse_action_parameters(handler):
    parameters = [
        {
            'key': 'key1',
            'value': 'value1'
        },
        {
            'key': 'key2',
            'value': 'value2'
        }
    ]
    expected = {
        'key1': 'value1',
        'key2': 'value2'
    }
    assert handler._parse_action_parameters(parameters) == expected

def test_exception_during_event_handling(mocker, handler, create_event):
    event = create_event(event_type='MESSAGE', message_text='Hello')
    mocker.patch.object(HangoutsChatHandler, 'handle_message')
    mocker.patch.object(HangoutsChatHandler, 'handle_exception')
    handler.handle_message.side_effect = Exception
    handler.handle_chat_event(event)
    handler.handle_exception.call_count == 1




import os
import json

import google_auth_httplib2
from googleapiclient import discovery
from google.oauth2 import service_account


class HangoutsChatAPI:
    GOOGLE_CHAT_SCOPES = ['https://www.googleapis.com/auth/chat.bot']

    def __init__(self, service_account_info=None):
        self.api = self._initialize_api(service_account_info)

    def _initialize_api(self, service_account_info):
        if service_account_info is None and os.environ.get('GOOGLE_APPLICATION_CREDENTIALS'):
            with open(os.environ.get('GOOGLE_APPLICATION_CREDENTIALS')) as f:
                service_account_info = json.loads(f.read())
        credentials = service_account.Credentials.from_service_account_info(
            service_account_info, scopes=self.GOOGLE_CHAT_SCOPES)
        http = google_auth_httplib2.AuthorizedHttp(credentials)
        return discovery.build('chat', 'v1', http=http)

    def list_spaces(self, page_size=100):
        spaces_list = list()
        spaces = self.api.spaces()
        request = spaces.list(pageSize=page_size)
        while request is not None:
            response = request.execute()
            spaces_list += response.get('spaces', [])
            request = spaces.list_next(request, response)
        return spaces_list

    def get_space(self, name):
        return self.api.spaces().get(name=name).execute()

    def list_memberships(self, space_name, page_size=100):
        memberships = list()
        members = self.api.spaces().members()
        request = members.list(parent=space_name, pageSize=page_size)
        while request is not None:
            response = request.execute()
            memberships += response.get('memberships', [])
            request = members.list_next(request, response)
        return memberships

    def get_membership(self, name):
        return self.api.spaces().members().get(name=name).execute()

    def create_message(self, message, space_name, thread_id=None, thread_key=None):
        """ Sends an asynchronous message to Hangouts Chat. """
        # Update thread (will send as new message if thread_id is None)
        if thread_id is not None:
            message['thread'] = thread_id
        return self.api.spaces().messages().create(
            parent=space_name, body=message, threadKey=thread_key).execute()

    def get_message(self, name):
        return self.api.spaces().messages().get(name=name).execute()

    def delete_message(self, name):
        return self.api.spaces().messages().delete(name=name).execute()

    def update_message(self, name, message):
        update_kwargs = {
            'name': name,
            'body': message,
            'updateMask': 'text,cards'
        }
        return self.api.spaces().messages().update(**update_kwargs).execute()



from enum import Enum
from collections import Iterable


class ImageStyle(Enum):
    IMAGE = 'IMAGE'  # square
    AVATAR = 'AVATAR'  # circular


class Icon(Enum):
    AIRPLANE = 'AIRPLANE'
    BOOKMARK = 'BOOKMARK'
    BUS = 'BUS'
    CAR = 'CAR'
    CLOCK = 'CLOCK'
    CONFIRMATION_NUMBER_ICON = 'CONFIRMATION_NUMBER_ICON'
    DESCRIPTION = 'DESCRIPTION'
    DOLLAR = 'DOLLAR'
    EMAIL = 'EMAIL'
    EVENT_SEAT = 'EVENT_SEAT'
    FLIGHT_ARRIVAL = 'FLIGHT_ARRIVAL'
    FLIGHT_DEPARTURE = 'FLIGHT_DEPARTURE'
    HOTEL = 'HOTEL'
    HOTEL_ROOM_TYPE = 'HOTEL_ROOM_TYPE'
    INVITE = 'INVITE'
    MAP_PIN = 'MAP_PIN'
    MEMBERSHIP = 'MEMBERSHIP'
    MULTIPLE_PEOPLE = 'MULTIPLE_PEOPLE'
    PERSON = 'PERSON'
    PHONE = 'PHONE'
    RESTAURANT_ICON = 'RESTAURANT_ICON'
    SHOPPING_CART = 'SHOPPING_CART'
    STAR = 'STAR'
    STORE = 'STORE'
    TICKET = 'TICKET'
    TRAIN = 'TRAIN'
    VIDEO_CAMERA = 'VIDEO_CAMERA'
    VIDEO_PLAY = 'VIDEO_PLAY'


class ResponseType(Enum):
    TYPE_UNSPECIFIED = 'TYPE_UNSPECIFIED'
    NEW_MESSAGE = 'NEW_MESSAGE'
    UPDATE_MESSAGE = 'UPDATE_MESSAGE'
    REQUEST_CONFIG = 'REQUEST_CONFIG'


class OnClickMixin:
    def add_link(self, url):
        self.link = True
        self.link_url = url
        return self

    def add_action(self, action_method, parameters=None):
        self.action = True
        self.action_method = action_method
        self.action_parameters = parameters
        return self

    def _update_on_click(self, widget):
        link = getattr(self, 'link', False)
        action = getattr(self, 'action', False)
        if link:
            data = {
                'openLink': {
                    'url': self.link_url
                }}
        elif action:
            action_data = {'actionMethodName': self.action_method.value}
            if self.action_parameters:
                action_data.update({
                    'parameters': [{'key': k, 'value': v} for k, v in self.action_parameters.items()]
                })
            data = {'action': action_data}
        if link or action:
            on_click = {'onClick': data}
            widget.update(on_click)


class Message:
    ResponseType = ResponseType

    def __init__(self, *cards, **kwargs):
        self.cards = list(cards)
        self.response_type = kwargs.get('response_type')
        self.response_url = kwargs.get('response_url')
        self.text = kwargs.get('text')

    @classmethod
    def request_config(cls, url):
        return cls(
            response_type=ResponseType.REQUEST_CONFIG,
            response_url=url)

    def add_card(self, card):
        self.cards.append(card)

    def output(self):
        message = {}
        if self.cards:
            message.update({
                'cards': [c.output() for c in self.cards]
            })
        if self.text is not None:
            message.update({'text': self.text})
        if self.response_type is not None:
            response_type = {'type': self.response_type.value}
            if self.response_type == ResponseType.REQUEST_CONFIG:
                response_type.update({'url': self.response_url})
            message.update({'actionResponse': response_type})
        return message


class CardAction(OnClickMixin):
    def __init__(self, label):
        self.action_label = label

    def output(self):
        card_action = {'actionLabel': self.action_label}
        self._update_on_click(card_action)
        return card_action


class Card:
    def __init__(self, *components):
        self.card_actions = list()
        self.sections = list()
        self.header = None
        for component in components:
            if isinstance(component, CardHeader) and self.header is None:
                self.header = component
            elif isinstance(component, Section):
                self.sections.append(component)
            elif isinstance(component, CardAction):
                self.card_actions.append(component)

    def add_section(self, section):
        self.sections.append(section)

    def add_action(self, action):
        self.card_actions.append(action)

    def output(self):
        sections = [s.output() for s in self.sections]
        card = {'sections': sections}
        if self.header:
            header = self.header.output()
            card.update({'header': header})
        if self.card_actions:
            card_actions = [a.output() for a in self.card_actions]
            card.update({'cardActions': card_actions})
        return card


class Section:
    def __init__(self, *widgets):
        self.widgets = list(widgets)
        self.header = None
        if self.widgets and isinstance(self.widgets[0], str):
            self.header = self.widgets.pop(0)

    def add_widget(self, widget):
        self.widgets.append(widget)

    def output(self):
        widgets = [w.output() for w in self.widgets]
        section = {'widgets': widgets}
        if self.header is not None:
            section.update({'header': self.header})
        return section


class ButtonList:
    def __init__(self, *buttons):
        self.buttons = list(buttons)

    def add_button(self, button):
        self.buttons.append(button)

    def output(self):
        buttons = [b.output() for b in self.buttons]
        return {'buttons': buttons}


class CardHeader:
    ImageStyle = ImageStyle

    def __init__(self, title, subtitle, image_url=None, image_style=None):
        self.title = title
        self.subtitle = subtitle
        self.image_url = image_url
        self.image_style = image_style

    def output(self):
        header = {
            'title': self.title,
            'subtitle': self.subtitle
        }
        if self.image_url is not None:
            header.update({'imageUrl': self.image_url})
        if self.image_style is not None:
            header.update({'imageStyle': self.image_style.value})
        return header


class TextParagraph:
    def __init__(self, text):
        self.text = text

    def output(self):
        text_paragraph = {'text': self.text}
        return {'textParagraph': text_paragraph}


class KeyValue(OnClickMixin):
    Icon = Icon

    def __init__(self, content, top_label=None, bottom_label=None, icon=None, icon_url=None, button=None):
        self.top_label = top_label
        self.content = content
        self.bottom_label = bottom_label
        self.icon = icon
        self.icon_url = icon_url
        self.button = button

    def output(self):
        key_value = {'content': self.content}
        if '\n' in self.content:
            key_value.update({'contentMultiline': True})
        if self.top_label is not None:
            key_value.update({'topLabel': self.top_label})
        if self.bottom_label is not None:
            key_value.update({'bottomLabel': self.bottom_label})
        if self.button is not None:
            key_value.update({'button': self.button})
        if self.icon is not None:
            key_value.update({'icon': self.icon.value})
        elif self.icon_url is not None:
            key_value.update({'iconUrl': self.icon_url})
        self._update_on_click(key_value)
        return {'keyValue': key_value}


class Image(OnClickMixin):
    def __init__(self, image_url, aspect_ratio=None):
        self.image_url = image_url
        self.aspect_ratio = aspect_ratio

    def output(self):
        image = {'imageUrl': self.image_url}
        if self.aspect_ratio is not None:
            image.update({'aspectRatio': self.icon_url})
        self._update_on_click(image)
        return {'image': image}


class ImageButton(OnClickMixin):
    Icon = Icon

    def __init__(self, icon=None, icon_url=None, name=None):
        self.icon_url = icon_url
        self.icon = icon
        self.name = name

    def output(self):
        if self.icon is not None:
            button = {'icon': self.icon.value}
        elif self.icon_url is not None:
            button = {'iconUrl': self.icon_url}
        if self.name is not None:
            button = {'name': self.name}
        self._update_on_click(button)
        return {'imageButton': button}


class TextButton(OnClickMixin):
    Icon = Icon

    def __init__(self, text):
        self.text = text

    def output(self):
        button = {'text': self.text}
        self._update_on_click(button)
        return {'textButton': button}





import os
import pytest
from collections import OrderedDict

from hangouts_helper.message import (Message, Card, CardHeader, Section,
    Image, KeyValue, ButtonList, TextButton)


@pytest.fixture
def pizza_bot_message():
    test_dir = os.path.dirname(os.path.realpath(__file__))
    import json
    with open(os.path.join(test_dir, 'pizza_bot_example.json')) as f:
        example_json = f.read()
    return json.loads(example_json)

def test_text_message():
    message = Message(text='hello')
    assert message.output() == {'text': 'hello'}

def test_request_config():
    message = Message.request_config('https://example.com')
    expected = {
        'actionResponse': {
            'type': 'REQUEST_CONFIG',
            'url': 'https://example.com'
        }
    }
    assert message.output() == expected

def test_add_link():
    button = TextButton(text='OPEN ORDER').add_link(url='https://example.com/orders/...')
    expected = {
        'textButton': {
            'text': 'OPEN ORDER',
            'onClick': {
                'openLink': {
                    'url': 'https://example.com/orders/...'
                }
            }
        }
    }
    assert button.output() == expected

def test_add_action():
    from enum import Enum
    class ActionMethod(Enum):
        TEST_METHOD = 'TEST_METHOD'
    parameters = OrderedDict()
    parameters['param1'] = 'value1'
    parameters['param2'] = 'value2'
    button = TextButton(text='OPEN ORDER').add_action(
        action_method=ActionMethod.TEST_METHOD,
        parameters=parameters)
    expected = {
        'textButton': {
            'text': 'OPEN ORDER',
            'onClick': {
                'action': {
                    'actionMethodName': 'TEST_METHOD',
                    'parameters': [{
                        'key': 'param1',
                        'value': 'value1'
                    },
                    {
                        'key': 'param2',
                        'value': 'value2'
                    }]
                }
            }
        }
    }
    assert button.output() == expected

def test_section_with_header():
    section = Section('Header', Image('https://placeimg.com/640/480/animals'))
    expected = {
        'header': 'Header',
        'widgets': [{
            'image': {
                'imageUrl': 'https://placeimg.com/640/480/animals'
            }
        }]
    }
    assert section.output() == expected

def test_section_without_header():
    section = Section(Image('https://placeimg.com/640/480/animals'))
    expected = {
        'widgets': [{
            'image': {
                'imageUrl': 'https://placeimg.com/640/480/animals'
            }
        }]
    }
    assert section.output() == expected

def test_pizza_bot_full_example(pizza_bot_message):
    message = Message()
    message.add_card(
        Card(
            CardHeader(
                title='Pizza Bot Customer Support',
                subtitle='pizzabot@example.com',
                image_url='https://goo.gl/aeDtrS'),
            Section(
                KeyValue(top_label='Order No.', content='12345'),
                KeyValue(top_label='Status', content='In Delivery')),
            Section(
                'Location',
                Image(image_url='https://maps.googleapis.com/...')),
            Section(
                ButtonList(
                    TextButton(text='OPEN ORDER').add_link(url='https://example.com/orders/...')))))

    assert message.output() == pizza_bot_message



import logging
from enum import Enum


class EventType(Enum):
    ADDED_TO_SPACE = 'ADDED_TO_SPACE'
    MESSAGE = 'MESSAGE'
    CARD_CLICKED = 'CARD_CLICKED'
    REMOVED_FROM_SPACE = 'REMOVED_FROM_SPACE'


class SpaceType(Enum):
    ROOM = 'ROOM'
    DIRECT_MESSAGE = 'DM'


class UserType(Enum):
    HUMAN = 'HUMAN'
    BOT = 'BOT'


class HangoutsChatHandler:
    SpaceType = SpaceType
    EventType = EventType
    ActionMethod = Enum

    def __init__(self, logger=None, debug=False):
        if logger is None:
            logger = logging.getLogger(__name__)
        self.log = logger
        self.debug = debug

    def _parse_action_parameters(self, parameters):
        parsed_parameters = {}
        for p in parameters:
            key = p['key']
            value = p['value']
            parsed_parameters[key] = value
        return parsed_parameters

    def handle_chat_event(self, event, sent_asynchronously=False):
        try:
            self.sent_asynchronously = sent_asynchronously
            event_type = EventType(event['type'])
            space_type = SpaceType(event['space']['type'])
            if event_type == EventType.ADDED_TO_SPACE:
                return self.handle_added_to_space(space_type, event)
            elif event_type == EventType.REMOVED_FROM_SPACE:
                return self.handle_removed_from_space(space_type, event)
            elif event_type == EventType.CARD_CLICKED:
                action_method = self.ActionMethod(event['action']['actionMethodName'])
                action_parameters = self._parse_action_parameters(event['action']['parameters'])
                return self.handle_card_clicked(action_method, action_parameters, event)
            elif event_type == EventType.MESSAGE:
                message = event['message']['text']
                return self.handle_message(message, event)
        except Exception as e:
            logging.exception('Error handling chat event')
            return self.handle_exception(e, event)

    def handle_exception(self, e, event=None):
        if self.debug:
            return {'text': str(e)}

    def handle_added_to_space(self, space_type, event):
        pass

    def handle_message(self, message, event):
        pass

    def handle_card_clicked(self, action_method, action_parameters, event):
        pass

    def handle_removed_from_space(self, space_type, event):
        pass



from setuptools import setup, find_packages


LONG_DESCRIPTION = open('README.md').read()

INSTALL_REQUIRES = [
    'google-api-python-client',
    'google-auth-httplib2'
]


setup(
    name='hangouts-helper',
    use_scm_version=True,
    setup_requires=['setuptools_scm'],
    description='Helper Python classes for handling and responding to Hangouts Chat events',
    long_description=LONG_DESCRIPTION,
    long_description_content_type='text/markdown',
    url='http://github.com/christippett/hangouts-helper',
    author='Chris Tippett',
    author_email='c.tippett@gmail.com',
    license='MIT',
    package_dir={'': 'src'},
    packages=find_packages('src'),
    install_requires=INSTALL_REQUIRES,
    zip_safe=False,
    classifiers=[
        'Environment :: Web Environment',
        'Framework :: Django',
        'Operating System :: OS Independent',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7'
    ]
)
